import argparse

parser = argparse.ArgumentParser(description='Emissary')
parser.add_argument('uba_greta_emission_db_path', type=str)
parser.add_argument('wrfinput_file_path', type=str, help="wrfinput file path")
parser.add_argument('wrfchemi_file_path', type=str, help="wrfchemi file path")
parser.add_argument('output_file_path',  type=str)
parser.add_argument('--grid_mappings_path', type=str, default=None)
parser.add_argument('--emission_data_tables_path', type=str, default=None)

args = parser.parse_args()

# Initialize drivers

import netCDF4
import numpy as np
from tools import gridFromWrfInput, netCDF4FromWrfInputAndWrfChemi, status

from osgeo import ogr, osr
import os
import pickle
import datetime

fgdb_driver     = ogr.GetDriverByName("OpenFileGDB")
mem_driver      = ogr.GetDriverByName('MEMORY')

# Open files
fgdb            = fgdb_driver.Open(args.uba_greta_emission_db_path)
area_emissions  = fgdb.GetLayerByName('RASTER_EMI_Teil1')
point_emissions = fgdb.GetLayerByName('F_PRTR_RESULT_Teil1')

from collections import deque

# save attributes into dict for faster access
if args.emission_data_tables_path is None:
    args.emission_data_tables_path = os.path.dirname( args.uba_greta_emission_db_path )
area_emissions_data_pickle = os.path.join( args.emission_data_tables_path, os.path.basename( args.uba_greta_emission_db_path ) + "_AREA_DATA.pickle" )
if not os.path.exists( area_emissions_data_pickle ):
    area_emissions_data_raw = np.genfromtxt(args.uba_greta_emission_db_path + "_RASTER_EMI_Teil1.csv", delimiter=";", names=True)
    area_emissions_data = {}

    i       = 0
    tstat   = deque( [ datetime.datetime.now() ] )
    n       = len(area_emissions_data_raw)

    for linenum in range(n):
        i, tstat = status("Area emissions data", i, n, tstat)

        line                        = area_emissions_data_raw[linenum]
        oid                         = line['OBJECTID']
        area_emissions_data[oid]    = {}
        
        for name in [ l for l in line.dtype.names if not l == 'OBJECTID' ]:
            area_emissions_data[oid][name] = line[name]

    with open(area_emissions_data_pickle, "wb") as f:
        pickle.dump(area_emissions_data, file=f)

with open(area_emissions_data_pickle, "rb") as f:
    area_emissions_data = pickle.load(f)

if args.emission_data_tables_path is None:
    args.emission_data_tables_path = os.path.dirname( args.uba_greta_emission_db_path )
point_emissions_data_pickle = os.path.join( args.emission_data_tables_path, os.path.basename( args.uba_greta_emission_db_path ) + "_POINT_DATA.pickle" )
if not os.path.exists( point_emissions_data_pickle ):
    point_emissions_data_raw = np.genfromtxt(args.uba_greta_emission_db_path + "_F_PRTR_RESULT_Teil1.csv", delimiter=";", names=True)
    point_emissions_data = {}

    i       = 0
    tstat   = deque( [ datetime.datetime.now() ] )
    n       = len(point_emissions_data_raw)

    for linenum in range(n):
        i, tstat = status("Point emissions data", i, n, tstat)

        line                        = point_emissions_data_raw[linenum]
        oid                         = line['OBJECTID']
        point_emissions_data[oid]    = {}
        
        for name in [ l for l in line.dtype.names if not l == 'OBJECTID' ]:
            point_emissions_data[oid][name] = line[name]

    with open(point_emissions_data_pickle, "wb") as f:
        pickle.dump(point_emissions_data, file=f)

with open(point_emissions_data_pickle, "rb") as f:
    point_emissions_data = pickle.load(f)


# figure out highest emission source:
max_emiss_height = np.max([ feature['emission_height'] for feature in point_emissions ])
point_emissions.ResetReading()

def find_level_in_wrfinput(z, nc, x, y):
    zbot = nc.variables["ZBOTTOM"][0,:,y,x]
    ztop = nc.variables["ZTOP"][0,:,y,x]
#
    level = -1
    for bot, top, lev in zip( zbot, ztop, range(len(zbot)) ):
        if bot <= z and top > z:
            level = lev
            break
#
    return level

grid_shp        = mem_driver.CreateDataSource('memData')
tmp             = mem_driver.Open('memData',1)

output_grid     = gridFromWrfInput(grid_shp, args.wrfinput_file_path)
nc              = netCDF4FromWrfInputAndWrfChemi(args.wrfinput_file_path, args.wrfchemi_file_path, args.output_file_path, zmax=max_emiss_height)

output_variables       = [ x for x in nc.variables if x.startswith("E_") ]

output_nox_variables   = [ "E_NO", "E_NO2" ]
output_voc_variables   = [ "E_BIGALK", "E_BIGENE", "E_C2H4", "E_C2H5OH", "E_C2H6", "E_CH2O", "E_CH3CHO", "E_CH3COCH3", "E_CH3OH", "E_MEK", "E_TOLUENE", "E_C3H6", "E_C3H8", "E_ISOP", "E_C10H16", "E_C2H2", "E_BENZENE", "E_XYLENE", "E_GLY", "E_MACR", "E_MGLY", "E_MVK", "E_HCOOH" ]
output_pm25_variables  = [ "E_ECI", "E_ECJ", "E_ORGI", "E_ORGJ", "E_PM25I", "E_PM25J", "E_SO4I", "E_SO4J", "E_NO3I", "E_NO3J", "E_NH4I", "E_NH4J", "E_NAI", "E_NAJ", "E_CLI", "E_CLJ" ]

#"E_CO_A", "E_CO_BB", "E_ORGI_A", "E_ORGI_BB", "E_ORGJ_A", "E_ORGJ_BB"

isnox   = lambda spec : spec in output_nox_variables
isvoc   = lambda spec : spec in output_voc_variables
ispm25  = lambda spec : spec in output_pm25_variables

def translate_output_to_input_species(name):
    testname = name
    if isnox(testname):
        testname = "E_NOx" 
    if isvoc(testname):
        testname = "E_NMVOC" 
    if ispm25(testname):
        testname = "E_PM2_5" 
#
    translation_table = {
        "E_NMVOC": "E_SUM_NMVOC",
        "E_CO":    "E_SUM_CO",
        "E_NH3":   "E_SUM_NH3",
        "E_NOx":   "E_SUM_NOX",
        "E_SO2":   "E_SUM_SO2",
        "E_PM_10": "E_SUM_PM10",
        "E_PM2_5": "E_SUM_PM2_5"
    }
    if testname in translation_table:
        return translation_table[testname]
    else:
        return None

# ---

# set up transformation between datasets
wgs_srs = osr.SpatialReference()
wgs_srs.ImportFromEPSG(4326)

transform_output_grid_to_emiss   = osr.CoordinateTransformation(output_grid.GetSpatialRef(), area_emissions.GetSpatialRef())
transform_emiss_to_output_grid   = osr.CoordinateTransformation(area_emissions.GetSpatialRef(), output_grid.GetSpatialRef())
transform_output_grid_to_wgs84   = osr.CoordinateTransformation(output_grid.GetSpatialRef(), wgs_srs)

# ---

def unit_conversion_factor_GRETA_to_WRF(species, area):
    # gases     in GRETA: kt a^-1  in WRF: mol km^-2 hr^-1
    # particles in GRETA: kt a^-1  in WRF: ug m^-2 s^-1
    mws_greta = {
        "E_SUM_NMVOC": 12.01e-3,
        "E_SUM_CO":    28.01e-3,
        "E_SUM_NH3":   17.03e-3,
        "E_SUM_NOX":   46.05e-3,
        "E_SUM_SO2":   64.06e-3
    }

    fac = -1.0
    if species in [ "E_SUM_NMVOC", "E_SUM_CO", "E_SUM_NH3", "E_SUM_NOX", "E_SUM_SO2" ]:
        # for gases: kt -> kg, kg -> mol, mol km^-2,   a^-1 -> h^-1
        fac =      1e6   * 1.0 / mws_greta[species] * 1e6 / area * 1.0 / (24*365)
    elif species in [ "E_SUM_PM10", "E_SUM_PM2_5" ]:
        # for particles: kt -> ug,    ug m^-2,   a^-1 -> s^-1
        fac = 1e15  *  1.0 / area * 1.0 / (24*365*86400)
    else:
        print("UNKNOWN species {:s}".format(species))
    
    return fac

# ---

def add_emissions(grid_data, t, z, y, x, input_oid, grid_cell_area, input_data, greta_area_fraction, wrf_area_fraction, needToSumInputVariables=False, cumulative=False):
#
    # loop through all possible emission output variables
    for output_variable in output_variables:
        # see if we have input
        input_variable = translate_output_to_input_species(output_variable)
#
        if not input_variable is None:
#            print("Translating {:s} into {:s}".format(input_variable, output_variable))

            speciation_fraction = 1.0
#
            existing_value = grid_data[output_variable][t,z,y,x]
#
            group_variables = [ output_variable ]
#
            if isnox(output_variable):
                group_variables = output_nox_variables
#
            if isvoc(output_variable):
                group_variables = output_voc_variables
#
            if ispm25(output_variable):
                group_variables = output_pm25_variables
#
            # always use ground level for speciation (also for elevated point sources...)
            group_total         = np.sum( [ grid_data[group_variable][t,0,y,x] for group_variable in group_variables ] )
#
            speciation_fraction = 1.0
            if group_total > 0.0:
                # should be 1.0 if not isnox or isvoc or ispm25 ...
                speciation_fraction = grid_data[output_variable][t,0,y,x] / group_total
#
            def get_input_value(input_variable, speciation_fraction, needToSumInputVariables):
                input_variables_raw = [ input_variable ]
                if needToSumInputVariables:
                    input_variables_raw = [ field for field in input_data[input_oid].keys() if field.endswith(input_variable.replace("E_SUM", "")) ]
#
                return np.sum( [ speciation_fraction * input_data[input_oid][ivar] for ivar in input_variables_raw ] )
#
            input_value = get_input_value( input_variable, speciation_fraction, needToSumInputVariables)
#
            if input_variable == "E_SUM_PM10":
                # PM 10 is actually the difference between PM 10 and PM 2.5 ...
                pm25_value = get_input_value( "E_SUM_PM2_5", speciation_fraction, needToSumInputVariables)
                input_value -= pm25_value
            
            input_value    = unit_conversion_factor_GRETA_to_WRF(input_variable, grid_cell_area) * input_value * greta_area_fraction
#
            if cumulative:
                # point sources are not (fractionally) replacing existing data, but add up
                grid_data[output_variable][t,z,y,x] += input_value
            else:
                # area sources replace existing values (fractionally)
                grid_data[output_variable][t,z,y,x]  = (1.0 - wrf_area_fraction) * existing_value + wrf_area_fraction * input_value * 100.0
#        else:
#            print("Cannot translate output species: {:s}".format(output_variable))

    grid_data['GRETA_fraction'][t, z, y, x] += wrf_area_fraction

    return grid_data

# area sources, to be distributed at the ground (lowest) level

if args.grid_mappings_path is None:
    args.grid_mappings_path = os.path.dirname( args.wrfchemi_file_path )
area_grid_mapping_pickle = os.path.join( args.grid_mappings_path, os.path.basename( args.uba_greta_emission_db_path ) + "_" + os.path.basename( args.wrfchemi_file_path ) + "_AREA_grid_mapping.pickle" )
if not os.path.exists( area_grid_mapping_pickle ):

    i       = 0
    tstat   = deque( [ datetime.datetime.now() ] )
    n       = output_grid.GetFeatureCount()

    area_gridcell_mapping = {}

    for output_gridcell in output_grid:
        i, tstat = status("Area grid mapping", i, n, tstat)
    #
        output_gridcell_geometry = output_gridcell.geometry().Clone()
        output_gridcell_geometry.Transform(transform_output_grid_to_emiss) # output_gridcell now in Emiss coords

        output_gridcell_area = output_gridcell_geometry.GetArea()
    #
        # only look at those features that intersect with geom
        area_emissions.SetSpatialFilter(output_gridcell_geometry)
    #
        y = output_gridcell.GetField("y"); x = output_gridcell.GetField("x")
    #
        output_fid = output_gridcell.GetFID()
        area_gridcell_mapping[output_fid] = { "x": x, "y": y, "z": 0, "area": output_gridcell_area, "input": {} }

        for emiss_gridcell in area_emissions:
            emiss_gridcell_geometry = emiss_gridcell.geometry()

            intersection        = output_gridcell_geometry.Intersection(emiss_gridcell_geometry)
            intersection_area   = intersection.GetArea()

            # (1) fraction of the GRETA grid cell in this WRF grid cell
            #     --> emitted amounts are to be reduced to this fraction of the total
            greta_area_fraction = intersection_area / emiss_gridcell_geometry.GetArea()
            # (2) fraction of the WRF grid cell covered by the intersecting part of the emissions grid cell
            #     --> this determines the relative contributions of existing emissions and the new stuff coming here
            wrf_area_fraction   = intersection_area / output_gridcell_area
    #
            emiss_oid = emiss_gridcell.GetFID()
            area_gridcell_mapping[output_fid]["input"][emiss_oid] = { "greta_area_fraction": greta_area_fraction, "wrf_area_fraction": wrf_area_fraction }
    #    
        area_emissions.ResetReading()

    output_grid.ResetReading()

    with open(area_grid_mapping_pickle, "wb") as f:
        pickle.dump(area_gridcell_mapping, file=f)

with open(area_grid_mapping_pickle, "rb") as f:
    area_gridcell_mapping = pickle.load(f)



# point sources, elevated
if args.grid_mappings_path is None:
    args.grid_mappings_path = os.path.dirname( args.wrfchemi_file_path )
point_grid_mapping_pickle = os.path.join( args.grid_mappings_path, os.path.basename( args.uba_greta_emission_db_path ) + "_" + os.path.basename( args.wrfchemi_file_path ) + "_POINT_grid_mapping.pickle" )
if not os.path.exists( point_grid_mapping_pickle ):

    i       = 0
    tstat   = deque( [ datetime.datetime.now() ] )
    n       = point_emissions.GetFeatureCount()

    point_gridcell_mapping = {}

    for point in point_emissions:
        i, tstat = status("Point grid mapping", i, n, tstat)
    #
        geom = point.geometry().Clone()
        geom.Transform(transform_emiss_to_output_grid) # point now in grid coords
    #
        # find the feature that contains the point (should be only one, the target output_grid output_gridcell)
        cellCount=0
        for output_gridcell in output_grid:
            output_gridcell_geometry = output_gridcell.geometry()
            if geom.Within(output_gridcell_geometry):
#
                t = 0; z = 0; y = output_gridcell.GetField("y"); x = output_gridcell.GetField("x")
                z = find_level_in_wrfinput( point['emission_height'], nc, x, y )
#
                output_fid = output_gridcell.GetFID()
                emiss_oid = point.GetFID()
                point_gridcell_mapping[output_fid] = { "x": x, "y": y, "z": z, "area": output_gridcell_geometry.GetArea(), 
                                                        "input": { emiss_oid: { "greta_area_fraction": 1.0, "wrf_area_fraction": 1.0 }} }
#
                cellCount += 1
#
        if cellCount  > 1:
            print("Something fishy with feature {:s}, multiple assignments!".format(point['plant_name']))
            exit
#
        output_grid.ResetReading()

    with open(point_grid_mapping_pickle, "wb") as f:
        pickle.dump(point_gridcell_mapping, file=f)

with open(point_grid_mapping_pickle, "rb") as f:
    point_gridcell_mapping = pickle.load(f)







# now, do the actual data replacements

grid_data = { name: nc[name][:] for name in nc.variables if name.startswith("E_") }
grid_data['GRETA_fraction'] = np.zeros_like(nc['E_NO2'][:])

i       = 0
tstat   = deque( [ datetime.datetime.now() ] )
n       = len(area_gridcell_mapping)

for output_cellmapping in area_gridcell_mapping.values():
    i, tstat = status("Area sources", i, n, tstat)

    t = 0; z = 0; y = output_cellmapping["y"]; x = output_cellmapping["x"]
#
    for emiss_oid, input_relations in output_cellmapping["input"].items():
        grid_data = add_emissions(grid_data, t, z, y, x, emiss_oid, output_cellmapping["area"], area_emissions_data, input_relations["greta_area_fraction"], input_relations["wrf_area_fraction"])

# point sources, elevated

i       = 0
tstat   = deque( [ datetime.datetime.now() ] )
n       = point_emissions.GetFeatureCount()

for output_cellmapping in point_gridcell_mapping.values():
    i, tstat = status("Point sources", i, n, tstat)

    t = 0; z = output_cellmapping["z"]; y = output_cellmapping["y"]; x = output_cellmapping["x"]
#
    for emiss_oid, input_relations in output_cellmapping["input"].items():
        grid_data = add_emissions(grid_data, t, z, y, x, emiss_oid, output_cellmapping["area"], point_emissions_data, input_relations["greta_area_fraction"], input_relations["wrf_area_fraction"], needToSumInputVariables=True, cumulative=True)


for name in nc.variables:
    if name.startswith("E_"):
        nc[name][:] = grid_data[name][:]

nc['GRETA_fraction'][:] = grid_data['GRETA_fraction'][:]

nc.close()
